// cbun: easy-going implementation of deterministric bunsetsu chunker
//  $Id: cbun.cpp 575 2011-08-29 13:19:52Z ynaga $
// Copyright (c) 2010-2011 Naoki Yoshinaga <ynaga@tkl.iis.u-tokyo.ac.jp>
#include <cstdio>
#include <vector>
#include <string>
#include "typedef.h"
#include "cedar.h"
#include "pa.h"    // opal
#include "polyk.h" // pecco

//  - IN/OUT format: MeCab/KNP; assuming JUMAN POS; tweak field_t accordingly
//    - labeled data: mark '* ' with the beginning morph of each bunsetsu
//  - classifier:
//    - training: opal (given options are delegated to opal)
//    - testing:  opal (Kd=0) / pecco (otherwise; modify PECCO_OPT)
//  - performance (on Intel 3.2Ghz CPU; opal options: Kd=2, algo=PA1, ave=1, s=1, c=0.005):
//    + training time: ~15s
//    + chunking accuracy: ~99.7%, compalable with KNP or CRF (w/ same feature)
//    + chunking speed:    ~20000 sent./sec. (almost same as MeCab)

namespace cbun {
  // default parameters
  static const char * PECCO_OPT = "-s 0 -r 0 -c fst -z 5 -zt nff";
  enum field_t {SURF, CPOS, FPOS, TYPE, INFL, FIN, YOMI, OTHER, NUM_FIELD};
  
  struct node {
    int      y;
    int      field[NUM_FIELD];
    ny::uint ilen;
    char     istr[1024];       // output
    node (int y_ = -1) : y (y_), ilen (0) {};
  };
  typedef cedar::da <int, -1, -2> sbag; // feature bag
  class chunker {
  private:
    const opal::option    &_opt;
    opal::Model*          _learner;
    opal::mem_pool        _ex_pool;
    std::vector <node>    _s;          // sentence
    size_t                _slen;       // # morphs
    ny::fv_t              _fv;         // feature vector
    opal::ex_t            _x;
    sbag                  _sbag;       // string bag: string -> int
    ny::uint              _nf;
    std::string           _dfn;        // trie
    std::string           _efn;        // event file
    FILE*                 _writer;
    ny::uint _pp, _pn, _np, _nn;
    size_t _ns;
    
    template <opal::mode_t MODE>
    void _set_node (char* cs, const size_t &len, int &y) {
      if (_slen >= _s.size ()) _s.resize (_slen + 1);
      node &n = _s[_slen];
      n.y = y;
      y = -1; // reset
      ny::uint i = 0;
      const char* const pend = cs + len;
      char* q[NUM_FIELD];
      for (char* p = cs; cs < pend && i < OTHER; ++i) {
        p = ny::skipChar (p, i == SURF ? '\t' : ',', pend);
        if (MODE == opal::TEST) q[i] = cs;
        if (i != TYPE && i < FIN) {
          int id = 0;
          if (MODE == opal::TRAIN) { // TRAIN
            int &l = _sbag.update (cs, p - cs - 1);
            id = l ? l : l = ++_nf;  // register
          } else {
            id = _sbag.exactMatchSearch <sbag::result_type> (cs, p - cs - 1);
          }
          n.field[i] = id * NUM_FIELD + i;
        }
        cs = p;
      }
      if (MODE == opal::TEST) // most time consuming part
        n.ilen = std::sprintf (&n.istr[0], "%s %s %s %s %s %s %s %s\n",
                               q[SURF], q[YOMI], std::strcmp (q[SURF], q[FIN]) ? q[FIN] : "*",
                               q[CPOS], q[FPOS], q[TYPE], q[INFL], cs);
      ++_slen;
    }
    void _set_feature (ny::uint i, bool sort) {
      _fv.clear ();
      for (ny::uint j = 0; j < FIN; ++j) { // NUM_FIELD
        if (j == TYPE) continue;
        if (i >= 2 && _s[i-2].field[j] > 0) _fv.push_back (_s[i-2].field[j] * 3);
        if (_s[i-1].field[j] > 0) _fv.push_back (_s[i-1].field[j] * 3 + 1);
        if (_s[i].field[j]   > 0) _fv.push_back (_s[i].field[j]   * 3 + 2);
      }
      if (sort) std::sort (_fv.begin (), _fv.end ());
    }
    void _print_ex () const {
      std::fwrite (_x.getLabel () > 0 ? "+1" : "-1", 1, 2, _writer);
      for (opal::fv_it it = _x.begin (); it != _x.end (); ++it)
        std::fprintf (_writer, " %d:1", *it);
      std::fwrite ("\n", 1, 1, _writer);
    }
    template <opal::mode_t MODE, typename classifier>
    void _process (classifier* p = 0, bool output = false) {
      ny::uint bid = 0;
      if (MODE == opal::TEST) {
        ++_pp;
        if (output) {
          std::fprintf (stdout, "# S-ID:%ld\n", ++_ns);
          std::fwrite ("* 0 -1D\n", 1, 8, stdout);
          std::fwrite (_s[0].istr, 1, _s[0].ilen, stdout);
        }
      }
      for (ny::uint i = 1; i < _slen; ++i) {
        const node &n  = _s[i];
        _set_feature (i, MODE == opal::TRAIN);
        if (MODE == opal::TRAIN) {
          _learner->set_ex (_x, n.y, _fv, true, _opt.d > 0);
          _ex_pool.put (_x);
          _print_ex ();
        } else {
          if (p->classify (_fv) >= 0) {
            if (output) std::fprintf (stdout, "* %d -1D\n", ++bid);
            if (n.y > 0) ++_pp; else ++_np;
          } else {
            if (n.y > 0) ++_pn; else ++_nn;
          }
          if (output) std::fwrite (n.istr, 1, n.ilen, stdout);
        }
      }
    }
    template <opal::mode_t MODE, typename classifier>
    void _analyze (const char* fn, classifier * p = 0) {
      std::string mode;
      switch (MODE) {
        case opal::TRAIN: mode = "learn";  break;
        case opal::TEST:  mode = "chunk"; break;
        default: break;
      }
      FILE* reader = std::strcmp (fn, "_") == 0 ? stdin : std::fopen (fn, "r");
      if (! reader) ny::print_err (HERE "no such file: %s\n", fn);
      if (reader == stdin)
        std::cerr << mode << " (input: STDIN)" << std::endl;
      bool output = MODE == opal::TEST && (reader == stdin || _opt.debug == 1);
      char buf[1 << 20]; std::setvbuf (reader, buf, _IOFBF, 1 << 20);
      int      y    = +1;
      ny::uint n    = 0;
      char*    line = 0;
      size_t   read = 0;
      while (ny::getLine (reader, line, read)) { // chunk bunsetsu
        char *str = line;
        switch (*line) {
          case 'E': // EOS
            _process <MODE> (p, output);
            _slen = 0;
            if (MODE != opal::TEST)
              if (++n % 1000 == 0)
                std::cerr << "\r" << mode << ": " << n << " sent. processed";
          case '#': // comment
            if (MODE == opal::TEST) {
              if (output) {
                std::fwrite (str, 1, read - 1, stdout);
                std::fwrite ("\n", 1, 1, stdout);
                if (*line == 'E') std::fflush (stdout);
              }
            }
            break;
          case '*': y = +1; str += 2; read -= 2; // skip mark ' *'
          default:
            *(str + read - 1) = '\0';
            _set_node <MODE> (str, read - 1, y); break;
        }
      }
      if (reader != stdin) std::fclose (reader);
      if (MODE != opal::TEST)
        std::cerr << "\r" << mode << ": " << n << " sent. processed; done." << std::endl;
    }
  public:
    chunker (const opal::option &opt_) : _opt (opt_), _learner (0), _s (), _slen (0), _fv (), _x (), _sbag (), _nf (0), _dfn (std::string (_opt.model) + ".cda"), _efn (std::string (_opt.model) + ".event"), _writer (0), _pp (0), _pn (0), _np (0), _nn (0), _ns (0) {};
    ~chunker () {}
    void train () {
      _learner = new opal::Model (_opt.d, _opt.N, _opt.algo, _opt.ave, _opt.s, _opt.c, _opt.M, _opt.device);
      _writer = std::fopen (_efn.c_str (), "w");
      if (! _writer) {
        std::fprintf (stderr, "cannot write to file: %s\n", _efn.c_str ());
        std::exit (1);
      }
      _analyze <opal::TRAIN> (_opt.train, _learner);
      std::fclose (_writer);
      _learner->train (_ex_pool, _opt.iter);
      _learner->save (_opt.model);
      delete _learner;
      _sbag.save (_dfn.c_str ());
    }
    void test (char * pecco_opt) {
      _sbag.open (_dfn.c_str ());
      if (_opt.d == 0) {
        opal::Model * _classifier = new opal::Model (_opt.d, _opt.N);
        _classifier->load (_opt.model);
        _analyze <opal::TEST> (_opt.test, _classifier);
        delete _classifier;
      } else { // delegate polynomial kernel computation to pecco
        char opt[256];
        std::sprintf (opt, "%%0 -m %s -lf %s -cf %s %s", _opt.model, _efn.c_str (), _efn.c_str (), pecco_opt ? pecco_opt : PECCO_OPT);
        pecco::PolyK * _classifier = new pecco::PolyK (opt);
        _classifier->load ();
        _analyze <opal::TEST> (_opt.test, _classifier);
        delete _classifier;
      }
      std::fprintf (stderr, "acc. %2.3f%% (pp %d) (pn %d) (np %d) (nn %d)\n",
                    (_pp+_nn) * 100.0 / (_pp+_pn+_np+_nn), _pp, _pn, _np, _nn);
    }
  };
}

int main (int argc, char** argv) {
  int    pecco_clen = 0;
  int    pecco_argc = argc;
  char * pecco_opt  = 0;
  while (--pecco_argc) {
    if (std::strcmp (argv[pecco_argc], "--") == 0) { // pecco opt overloaded
      argc = pecco_argc;
      pecco_opt = new char[pecco_clen];
      for (int i = 0; i < pecco_clen;)
        i += std::sprintf (&pecco_opt[i], " %s", argv[++pecco_argc]);
      break;
    }
    pecco_clen += std::strlen (argv[pecco_argc]) + 1;
  }
  opal::option opt (argc, argv);
  cbun::chunker c (opt);
  if (opt.mode != opal::TEST)  c.train ();
  if (opt.mode != opal::TRAIN) c.test  (pecco_opt);
  if (pecco_opt) delete [] pecco_opt;
  return 0;
}
